#ifndef ross_infinite_tape_header
#define ross_infinite_tape_header

/*Authored 2002    by  Kyle  Ross,  Undergraduate  Philosophy   Thesis
(Advisor: Professor Bram van Heuveln)

Modified 2003    by Owen Kellett, Undergraduate Research
(Advisor: Professor Bram van Heuveln)

This  file is part of the   "Generic Turing Machine Implementation for
Use in the Exploration of the Busy Beaver Problem".

The "Generic Turing Machine Implementation for  Use in the Exploration
of the Busy Beaver Problem" is free  software; you can redistribute it
and/or modify it under the terms of the  GNU General Public License as
published   by the Free Software Foundation;   either version 2 of the
License, or (at your option) any later version.

This programme is distributed in the hope that it  will be useful, but
WITHOUT    ANY  WARRANTY;  without   even  the    implied warranty  of
MERCHANTABILITY or FITNESS FOR  A  PARTICULAR  PURPOSE.  See  the  GNU
General Public License for more details.

You should have received   a copy of  the  GNU General Public  License
along  with  this   programme; if  not,  write   to the Free  Software
Foundation, Inc., 59  Temple Place, Suite  330, Boston, MA  02111-1307
USA (or visit http://www.fsf.org/licenses/licenses.html)

Created as part of the Busy Beaver Project:
Bram van Heuveln <heuveb@rpi.edu>
Boleshaw Szymanski <szymansk@cs.rpi.edu>
Selmer Bringsjord <selmer@rpi.edu>
Carlos Varela <cvarela@cs.rpi.edu>
Owen Kellett <kelleo@rpi.edu>
Shailesh Kelkar <kelkas@rpi.edu>
Kyle Ross <rossk2@rpi.edu>
Rensselaer Reasoning Group
Department of Cognitive Science
Department of Computer Science
Rensselaer Polytechnic Institute
Troy, New York 12180
U.S.A.
*/
//implements an infinite-length tape of an arbitrary alphabet to be used in a generalised turing machine implementation; assumes small-size alphabet (this could easily be expanded through a derived class)
//enforcing of alphabet rules (i.e. which symbols belong to the alphabet, etc.) will be left to the finite-state automaton

#include <iostream>
#include <list>
#include "debug.hpp"

/**
 * @defgroup TuringMachineDefinition
 */

/**
 * an auto-expanding wrapper for an std::list.
 * provides constant amortised time operations
 * assumes small types for storage (e.g. int, bool, etc.) since parameters are passed by value
 * @ingroup TuringMachineDefinition
 */
template <typename T>
class infinite_tape
{
public:
	enum direction{left, right, do_nothing, halt_move};

	/**
	 * default Constructor.
	 * @author Owen Kellett
	 */
	infinite_tape()
	:tape(start_size, 0), default_value(0), counter(0), non_blank_count(0)
	{
		current=tape.begin();
		for(typename std::list<T>::size_type i=start_size/2; i!=0; --i)
		{
			++current;
		}
		cardinal=current;
		#ifdef ross_debug_mode
			std::cout<<"["<<this<<"]infinite-tape("<<0<<") constructed"<<std::endl;
		#endif
	}

	//complexity: O(1)+overhead for construction incurred by the implementation (O(1) since start_size is constant)
	/**
	 * construct a tape of start_size with each value set to default value.
	 * set current iterators to the middle element, start at this index, and counter to zero
	 * @param new_default_value default value for tape
	 * @author Kyle Ross
	 */
	infinite_tape(T new_default_value)
	:tape(start_size, new_default_value), default_value(new_default_value), counter(0), non_blank_count(0)
	{
		current=tape.begin();
		for(typename std::list<T>::size_type i=start_size/2; i!=0; --i)
		{
			++current;
		}
		cardinal=current;
		#ifdef ross_debug_mode
			std::cout<<"["<<this<<"]infinite-tape("<<default_value<<") constructed"<<std::endl;
		#endif
	}

	//complexity: O(new_size)+overhead for construction incurred by the implementation
	/**
	 * construct a tape of new_size with each value set to default_value.
	 * set current iterators to the middle element, start at this index, and counter to zero
	 * @param new_size initial size of tape
	 * @param new_default_value default value of tape
	 * @author Kyle Ross
	 */
	infinite_tape(typename std::list<T>::size_type new_size, T new_default_value)
	:tape(new_size, new_default_value), default_value(new_default_value), counter(0), non_blank_count(0)
	{
		current=tape.begin();
		for(typename std::list<T>::size_type i=new_size/2; i!=0; --i)
		{
			++current;
		}
		cardinal=current;
		#ifdef ross_debug_mode
			std::cout<<"["<<this<<"]infinite-tape("<<new_size<<", "<<default_value<<") constructed"<<std::endl;
		#endif
	}

	//complexity: O(N) in the allocated size of hrs+overhead for construction incurred by the implementation
	/**
	 * copy constructor.
	 * @pre rhs is a valid tape
	 * @post *this is a tape functionally equivalent to rhs
	 * @param rhs valid infinite_tape to copy
	 * @author Kyle Ross
	 */
	infinite_tape(const infinite_tape<T> &rhs)
	:tape(rhs.tape), default_value(rhs.default_value), counter(0), non_blank_count(rhs.non_blank_count)
	{
		current=tape.begin();
		typename std::list<T>::const_iterator matcher=rhs.tape.begin();
		while((matcher++)!=rhs.current)
		{
			++current;
		}
		cardinal=current;
	}

	/**
	 * assignment operator.
	 * @author Kyle Ross
	 */
	const infinite_tape<T>& operator=(const infinite_tape<T> &rhs)
	{
		tape=rhs.tape;
		current=cardinal=tape.begin();
		typename std::list<T>::const_iterator temp_iter=rhs.tape.begin();
		while(temp_iter!=rhs.current)
		{
			++current;
			++temp_iter;
		}
		temp_iter=rhs.tape.begin();
		while(temp_iter!=rhs.cardinal)
		{
			++cardinal;
			++temp_iter;
		}
		counter=rhs.counter;
		default_value=rhs.default_value;
		non_blank_count=rhs.non_blank_count;

		return *this;
	}

	/**
	 * trivial destructor
	 * @author Kyle Ross
	 */
	~infinite_tape()
	{
		#ifdef ross_debug_mode
			std::cout<<"["<<this<<"]infinite-tape destructed"<<std::endl;
		#endif
	}

	/**
	 * Read the value held by the cell under the read-head.
	 * @return the value held by the cell under the read-head
	 * @author Kyle Ross
	 */
	T read() const
	{
		return (*current);
	}

	/**
	 * Clears the tape
	 * @author Kyle Ross
	 */
	void clear()
	{
		*this=infinite_tape<T>(default_value);
	}

	/**
	 * Writes a value to the current location on the tape.
	 * @pre current refers to a valid cell of the tape
	 * @post to_write has been stored in the cell under the read-head
	 * @param to_write the symbol to write to the tape
	 */
	void write(T to_write)
	{
		//switching a blank to a non-blank increment non-blank count
		if(*current==default_value && to_write!=default_value)
		{
			++non_blank_count;
		}
		//switching a non-blank to a blank; decrement non-blank count
		else if(*current!=default_value && to_write==default_value)
		{
			--non_blank_count;
		}
		(*current)=to_write;
	}

	/**
	 * Move the head one position to the left xor to the right of the current position.
	 * maintains the invariant that the head iterator refers to a valid tape square
	 * makes sure that there is tape to the left of the current position to move to then moves
	 * @pre current refers to a valid location on the tape
	 * @post current refers to a valid location on the tape; if which==do_nothing then current remains unchanged; if which==left then current is one to the left of its previous position and if which==right then current is one to the right of its previous position, adding additional cells to the tape as necessary to maintain its theoretical infinity
	 * @param which direction to move the read head
	 * @author Kyle Ross
	 */
	void move(direction which)
	{
		if(which==do_nothing)
		{
			return;
		}
		tape_check(which);
		if(which==left)
		{
			#ifdef ross_debug_mode
				std::cout<<"["<<this<<"]moving left"<<std::endl;
			#endif
			--current;
		}
		else
		{
			#ifdef ross_debug_mode
				std::cout<<"["<<this<<"]moving right"<<std::endl;
			#endif
			++current;
		}
		++counter;
	}

	/**
	 * moves the read-head to the cardinal position.
	 * @author Kyle Ross
	 */
	void reset_head()
	{
		current=cardinal;
		#ifdef ross_debug_mode
			std::cout<<"["<<this<<"]resetting head to cardinal"<<std::endl;
		#endif
	}

	/**
	 * Checks if the tape is blank.
	 * @return true iff tape is blank else false
	 * @author Kyle Ross
	 */
	bool is_blank() const
	{
		return non_blank_count==0;
	}

	/**
	 * Are the symbols on the tape in the format (<blank>)*(<symbol>)*(<blank>)*?.
	 * @pre the tape has zero or more cells
	 * @post the function returns true iff the tape's symbols match the regexp above, false else; the minimal number of symbols required to determine the return-value are scanned
	 * @author Kyle Ross
	 */
	bool is_contiguous() const
	{
		typename std::list<T>::const_iterator iter=tape.begin();
		while(iter!=tape.end() && (*iter)==default_value)//recognises blanks at beginning of tape
		{
			++iter;
		}
		if(iter==tape.end())//an empty tape is contiguous? not sure how this should be defined
		{
			return true;
		}
		while(iter!=tape.end() && (*iter)!=default_value)//recognises non-blanks after blanks-block
		{
			++iter;
		}
		if(iter==tape.end())//blanks followed by symbols ... it's good
		{
			return true;
		}
		while(iter!=tape.end() && (*iter)==default_value)//recognises after symbols-block
		{
			++iter;
		}
		return iter==tape.end();//if all symbols of the tape have been recognised then it meets the contiguous requirement
	}

	/**
	 * Is the read-head on a blank?.
	 * @return true is returned iff the read-head is scanning a blank, false else
	 * @author Kyle Ross
	 */
	bool on_blank() const
	{
		return (*current)==default_value;
	}

	/**
	 * Are there any symbols to the <direction> of the read-head?.
	 * @pre current refers to a valid cell on the tape
	 * @post if which==do_nothing than false is vacuously returned; if which==left then each cell to the left of the read-head have been scanned and true is returned iff there exists at least one non-blank symbol in these cells, false else; if which==right then each cell to the right of the read-head have been scanned and true is returned iff there exists at least one non-blank symbol in these cells, false else
	 * @param which direction to check for symbols
	 * @author Kyle Ross
	 */
	bool symbols_to_direction(direction which) const
	{
		if(which==do_nothing)//meaningless
		{
			return false;
		}
		if(which==left)
		{
			typename std::list<T>::const_reverse_iterator iter(current);
			//no initial ++ because "the fundamental identity of reverse iterators is '&*(reverse_bidirectional_iterator(i)) == &*(i - 1)'" (quote from sgi-stl docs)
			while(iter!=tape.rend() && (*iter)==default_value)//move left until symbol reached xor beginning of tape reached
			{
				++iter;
			}
			return iter!=tape.rend();//if it's not reached the beginning of the tape, there must have been a symbol to the left of the read-head
		}
		else
		{
			typename std::list<T>::const_iterator iter(current);
			++iter;//don't consider symbol under read-head
			while(iter!=tape.end() && (*iter)==default_value)//move right until symbol reached xor end of tape reached
			{
				++iter;
			}
			return iter!=tape.end();//if it's not reached the end of the tape, there must have been a symbol to the right of the read-head
		}
	}

	/**
	 * Is the tape in standard position?.
	 * (not on_blank()) and is_contiguous() and (symbols_to_direction(left) xor symbols_to_direction(right)); the xor'ing is done by checking that the answers to "are there symbols to the left of the read head?" and "are there symbols to the right of the read head?" are different ... which ensures that there are symbols in exactly one direction and not the other ... sneaky, eh?
	 * @pre all requirements for is_blank, is_continuous, symbols_to(left), and symbols_to(right) are met
	 * @post true is returned iff the above conditional is satisfied, false else
	 * @author Kyle Ross
	 */
	bool in_standard_position() const
	{
		return (!on_blank() && is_contiguous() && (!symbols_to_direction(left) || !symbols_to_direction(right)));
	}

	/**
	 * How many occurrences are there of to_count on the tape?.
	 * @param to_count symbol to count on the tape
	 * @return number of occurences of to_count on the tape
	 * @author Kyle Ross
	 */
	typename std::list<T>::size_type count_occurrences(T to_count) const
	{
		typename std::list<T>::size_type accumulate(0);
		for(typename std::list<T>::const_iterator i=tape.begin(); i!=tape.end(); ++i)
		{
			if((*i)==to_count)
			{
				++accumulate;
			}
		}
		#ifdef ross_debug_mode
			if(to_count==default_value)
			{
				std::cout<<"["<<this<<"]count-occurrences WARNING count may be inaccurate since default_value was parameter"<<std::endl;
			}
		#endif
		return accumulate;
	}

	/**
	 * provides the old-style display using a very simple pre-defined inflexible format.
	 * @pre the tape contains at least one "cell" and current, cardinal are both valid positions on the tape
	 * @post if the tape is empty, the message "< tape is empty >" has been displayed, else, the symbols on the tape have been displayed, each followed by a single space character and if current==cardinal then "->[" has been displayed before and "]<-" after the head/cardinal, else, "->" is displayed before and "<-" after the head and "[" before and "]" after the cardinal
	 * @author Kyle Ross
	 */
	void show(std::ostream &out) const
	{
		if(tape.begin()==tape.end())
		{
			out<<"< tape is empty >";
		}
		else
		{
			for(typename std::list<T>::const_iterator i=tape.begin(); i!=tape.end(); ++i)
			{
				if(i==current)
				{
					out<<"->";
				}
				if(i==cardinal)
				{
					out<<"[";
				}
				out<<(*i);
				if(i==cardinal)
				{
					out<<"]";
				}
				if(i==current)
				{
					out<<"<-";
				}
				out<<' ';
			}
		}
	}

	/**
	 * wow, could it possibly get more complicated?. ... no, not complicated, *flexible*! ... each delimiter is of an arbitrary type (call it X_type) but must have an std::ostream& operator<<(ostream&, const X_type&) operator defined
	 * @pre the tape contains at least one "cell" and current, cardinal are both valid positions on the tape; all characters held on the tape are printable and out is in a valid output state
	 * @post the tape has been printed on out in a well-defined (but, due to its complexity, not explicated here) format
	 * @author Kyle Ross
	 */
	template <typename delim_type1, typename delim_type2, typename delim_type3, typename delim_type4, typename delim_type5, typename delim_type6, typename delim_type7, typename delim_type8>
	void show(std::ostream &out, const delim_type1 &before_delim, const delim_type2 &between_delim, const delim_type3 &after_delim, const delim_type4 &before_cardinal, const delim_type5 &after_cardinal, const delim_type6 &before_head, const delim_type7 &after_head, const delim_type8 &blank_delim, bool extra_between_delims) const
	{
		typename std::list<T>::size_type first_non_blank_index, last_non_blank_index, head_index, cardinal_index;
		bool symbols_on_tape(false);
		typename std::list<T>::const_iterator iter, first_non_blank, last_non_blank;

		//figures out the relative positions of and establishes and iterator to the first and last non-blank symbols, the head, and the cardinal
		typename std::list<T>::size_type index(0);
		for(iter=tape.begin(); iter!=tape.end(); ++iter, ++index)
		{
			if(iter==current)
			{
				head_index=index;
			}
			if(iter==cardinal)
			{
				cardinal_index=index;
			}
			if((*iter)!=default_value)
			{
				if(!symbols_on_tape)
				{
					first_non_blank=iter;
					first_non_blank_index=index;
					symbols_on_tape=true;
				}
				last_non_blank=iter;
				last_non_blank_index=index;
			}
		}
		//displays output for debugging the generalised display function
		#ifdef ross_debug_mode
			#ifdef ross_debug_verbose
				{
					typename std::list<T>::size_type i(0);
					for(iter=tape.begin(); iter!=tape.end(); ++iter, ++i)
					{
						if(i==head_index)
						{
							if(iter!=current)
							{
								std::cout<<"oh, no, head's index is set improperly-E01"<<std::endl;
								return;
							}
							std::cout<<"<head>";
						}
						if(i==cardinal_index)
						{
							if(iter!=cardinal)
							{
								std::cout<<"oh, no, current's index is set improperly-E02"<<std::endl;
								return;
							}
							std::cout<<"<cardinal>";
						}
						if(symbols_on_tape && i==first_non_blank_index)
						{
							if(iter!=first_non_blank)
							{
								std::cout<<"oh, no, fnb index is set improperly-E03"<<std::endl;
								return;
							}
							std::cout<<"<fnb>";
						}
						if(symbols_on_tape && i==last_non_blank_index)
						{
							if(iter!=last_non_blank)
							{
								std::cout<<"oh, no, lnb index is set improperly-E04"<<std::endl;
								return;
							}
							std::cout<<"<lnb>";
						}
						std::cout<<(*iter);
						if(symbols_on_tape && i==last_non_blank_index)
						{
							std::cout<<"<\\lnb>";
						}
						if(symbols_on_tape && i==first_non_blank_index)
						{
							std::cout<<"<\\fnb>";
						}
						if(i==cardinal_index)
						{
							std::cout<<"<\\cardinal>";
						}
						if(i==head_index)
						{
							std::cout<<"<\\head>";
						}
					}
					std::cout<<std::endl;
				}
			#endif
		#endif

		//adjust indices so that [first_non_blank_index, last_non_blank_index] demarks the set of symbols to be printed as proper symbols
		if(symbols_on_tape)
		{
			if(head_index<first_non_blank_index)
			{
				first_non_blank_index=head_index;
				first_non_blank=current;
			}
			else if(head_index>last_non_blank_index)
			{
				last_non_blank_index=head_index;
				last_non_blank=current;
			}
		}
		//displays output for debugging the generalised display function
		#ifdef ross_debug_mode
			#ifdef ross_debug_verbose
				{
					typename std::list<T>::size_type i(0);
					for(iter=tape.begin(); iter!=tape.end(); ++iter, ++i)
					{
						if(i==head_index)
						{
							if(iter!=current)
							{
								std::cout<<"oh, no, head's index is set improperly-E01"<<std::endl;
								return;
							}
							std::cout<<"<head>";
						}
						if(i==cardinal_index)
						{
							if(iter!=cardinal)
							{
								std::cout<<"oh, no, current's index is set improperly-E02"<<std::endl;
								return;
							}
							std::cout<<"<cardinal>";
						}
						if(symbols_on_tape && i==first_non_blank_index)
						{
							if(iter!=first_non_blank)
							{
								std::cout<<"oh, no, fnb index is set improperly-E03"<<std::endl;
								return;
							}
							std::cout<<"<fnb>";
						}
						if(symbols_on_tape && i==last_non_blank_index)
						{
							if(iter!=last_non_blank)
							{
								std::cout<<"oh, no, lnb index is set improperly-E04"<<std::endl;
								return;
							}
							std::cout<<"<lnb>";
						}
						std::cout<<(*iter);
						if(symbols_on_tape && i==last_non_blank_index)
						{
							std::cout<<"<\\lnb>";
						}
						if(symbols_on_tape && i==first_non_blank_index)
						{
							std::cout<<"<\\fnb>";
						}
						if(i==cardinal_index)
						{
							std::cout<<"<\\cardinal>";
						}
						if(i==head_index)
						{
							std::cout<<"<\\head>";
						}
					}
					std::cout<<std::endl;
				}
			#endif
		#endif

		//performs the output operations ... very complicated in format, but well-defined
		out<<before_delim;

		if(!symbols_on_tape) //if tape is empty, treat as special case
		{
			if(current==cardinal) //current and cardinal are same, singleton output
			{
				out<<before_cardinal<<before_head<<(*current)<<after_head<<after_cardinal;
			}
			else if(cardinal_index<head_index) //output elements [cardinal_index, head_index]
			{
				out<<before_cardinal<<blank_delim<<after_cardinal; //output cardinal_index, leaving (cardinal_index, head_index]
				if(extra_between_delims)
				{
					out<<between_delim;
				}
				for(typename std::list<T>::size_type i=cardinal_index+1; i<head_index-1; ++i) //output [cardinal_index+1, head_index-1), leaving [head_index-1, head_index]
				{
					out<<blank_delim<<between_delim;
				}
				if(head_index-cardinal_index>1)
				{
					out<<blank_delim; //output head_index-1, leaving head_index
					if(extra_between_delims)
					{
						out<<between_delim;
					}
				}
				out<<before_head<<(*current)<<after_head;
			}
			else //cardinal_index>head_index ... output elements [head_index, cardinal_index]
			{
				out<<before_head<<(*current)<<after_head; //output head_index, leaving [head_index+1, cardinal_index]
				if(extra_between_delims)
				{
					out<<between_delim;
				}
				for(typename std::list<T>::size_type i=head_index+1; i<cardinal_index-1; ++i) //output [head_index+1, cardinal_index-1)
				{
					out<<blank_delim<<between_delim;
				}
				if(cardinal_index-head_index>1)
				{
					out<<blank_delim; //output cardinal_index-1, leaving cardinal_index
					if(extra_between_delims)
					{
						out<<between_delim;
					}
				}
				out<<before_cardinal<<blank_delim<<after_cardinal;
			}
		}
		else
		{
			if(first_non_blank_index<=cardinal_index && cardinal_index<=last_non_blank_index) //cardinal lies within [fnb, lnb] so don't worry about it
			{
				typename std::list<T>::size_type i;
				for(iter=first_non_blank, i=first_non_blank_index; i<=last_non_blank_index; ++i, ++iter) //outputs [fnb, lnb]
				{
					if(i==cardinal_index) //cardinal tag (i.e. cardinal tag outside head tag)
					{
						out<<before_cardinal;
					}
					if(i==head_index) //head tag
					{
						out<<before_head;
					}
					out<<(*iter); //display symbol
					if(i==head_index) //head closing tag
					{
						out<<after_head;
					}
					if(i==cardinal_index) //cardinal closing tag
					{
						out<<after_cardinal;
					}
					if(i!=last_non_blank_index) //don't output between_delim after last index
					{
						if(i==cardinal_index || i==head_index || i==cardinal_index-1 || i==head_index-1) //only output between_delim before/after head/cardinal if that is required
						{
							if(extra_between_delims)
							{
								out<<between_delim;
							}
						}
						else
						{
							out<<between_delim;
						}
					}
				}
			}
			else if(cardinal_index<first_non_blank_index) //cardinal before [fnb, lnb], must print cardinal and blank place-holders
			{
				out<<before_cardinal<<blank_delim<<after_cardinal; //fnb<=head<=lnb, so cardinal can safely be assumed to *not* be the head; outputs [cardinal], leaving [cardinal+1, lnb]
				if(extra_between_delims)
				{
					out<<between_delim;
				}
				typename std::list<T>::size_type i;
				for(++(iter=cardinal), i=cardinal_index+1; i<first_non_blank_index; ++i, ++iter) //outputs [cardinal+1, fnb), leaving [fnb, lnb]
				{
					out<<blank_delim;
					if(i==head_index-1) //handle case where head immediately follows blanks; only print between_delim before if that is required
					{
						if(extra_between_delims)
						{
							out<<between_delim;
						}
					}
					else
					{
						out<<between_delim;
					}
				}
				for(iter=first_non_blank, i=first_non_blank_index; i<=last_non_blank_index; ++i, ++iter) //outputs [fnb, lnb]
				{
					if(i==head_index) //head tag
					{
						out<<before_head;
					}
					out<<(*iter); //output symbol
					if(i==head_index) //head closing tag
					{
						out<<after_head;
					}
					if(i!=last_non_blank_index) //do not output between_delim after last element
					{
						if(i==head_index || i==head_index-1) //do not output between_delim before/after head unless this is required
						{
							if(extra_between_delims)
							{
								out<<between_delim;
							}
						}
						else
						{
							out<<between_delim;
						}
					}
				}
			}
			else //cardinal after [fnb, lnb], again must print place-holder blanks and cardinal specially
			{
				typename std::list<T>::size_type i;
				for(iter=first_non_blank, i=first_non_blank_index; i<=last_non_blank_index; ++i, ++iter) //outputs [fnb, lnb], leaving [lnb+1, cardinal]
				{
					if(i==head_index) //head tag
					{
						out<<before_head;
					}
					out<<(*iter);
					if(i==head_index) //head closing tag
					{
						out<<after_head;
					}
					if(i!=last_non_blank_index) //don't output between_delim for last index
					{
						if(i==head_index || i==head_index-1 || i==cardinal_index-1) //don't output between_delim's before/after head unless required, use cardinal-1 check in case cardinal immediately follows lnb and don't output between_delim unless required
						{
							if(extra_between_delims)
							{
								out<<between_delim;
							}
						}
						else
						{
							out<<between_delim;
						}
					}
				}
				for(++(iter=last_non_blank), i=last_non_blank_index+1; i!=cardinal_index; ++i, ++iter) //outputs [lnb+1, cardinal), leaving cardinal
				{
					out<<blank_delim; //display place-holder blanks
					if(i==cardinal_index-1) //do not print between_delim before cardinal unless required
					{
						if(extra_between_delims)
						{
							out<<between_delim;
						}
					}
					else
					{
						out<<between_delim;
					}
				}
				out<<before_cardinal<<blank_delim<<after_cardinal; //outputs cardinal
			}
		}

		out<<after_delim;
	}

	//Begin Owen Kellett modifications
	//note: due to the nature of the non halting routines implemented, the following blatant breaches of information hiding were necessary

	/**
	 * Clears the tape to setting each location to the default value.
	 * @author Owen Kellett
	 */
	void reset()
	{
		for(typename std::list<T>::iterator i=tape.begin(); i!=tape.end(); ++i)
		{
			(*i) = default_value;
		}
		non_blank_count = 0;
		counter = 0;
	}

	/**
	 * Sets the read-head to location.
	 * @param location iterator of the location to set the read-head
	 * @author Owen Kellett
	 */
	void setCurrentLocation(typename std::list<T>::iterator location)
	{
		current = location;
	}

	/**
	 * default_value accessor.
	 * @return the default value used for the tape
	 * @author Owen Kellett
	 */
	T defaultValue()
	{
		return default_value;
	}

	/**
	 * Returns the read-head.
	 * @return an iterator pointing to the current tape position
	 * @author Owen Kellett
	 */
	typename std::list<T>::iterator currentLocation()
	{
		return current;
	}

	/**
	 * Returns the beginning of the tape.
	 * @return an iterator pointing to the beginning of the tape
	 * @author Owen Kellett
	 */
	typename std::list<T>::iterator begin()
	{
		return tape.begin();
	}

	/**
	 * Returns the beginning of the tape.
	 * @return a const_iterator pointing to the beginning of the tape
	 * @author Owen Kellett
	 */
	typename std::list<T>::const_iterator begin() const
	{
		return tape.begin();
	}

	/**
	 * Finds an iterator in the tape.
	 * used mainly for debugging purposes
	 * @param find iterator to search for
	 * @return true if find points to a location in the tape
	 * @author Owen Kellett
	 */
	bool find(typename std::list<T>::iterator find)
	{
		typename std::list<T>::iterator temp = tape.begin();
		while(temp != tape.end())
		{
			if(temp == find)
				return true;
			temp++;
		}
		return false;
	}

	/**
	 * tests for a simple loop repeat configuration in rhs.
	 * @pre this and rhs represent two points during a particular turing_machine execution; the current state at each point in the execution is the same; the current read head symbol at each point in the execution is the same;
	 * @post true if this tape and rhs having matching configurations that exhibit right motion simple loop behavior
	 * @param rhs representation of the tape at some point later on in the execution
	 * @param lhsLeftBound farthest cell to the left on the tape that the read head has been between the two points of execution
	 * @author Owen Kellett
	 */
	bool repeatConfigRight(infinite_tape<T> &rhs, typename std::list<T>::iterator lhsLeftBound)
	{
		typename std::list<T>::iterator lRightCounter = current;
		typename std::list<T>::iterator lLeftCounter = current;
		typename std::list<T>::iterator rRightCounter = rhs.current;
		typename std::list<T>::iterator rLeftCounter = rhs.current;
		bool go = false;

		//--check to the right of current
		//all cells starting from the read head of both states must be same
		//an unequal number of default_value's at the end is allowed (since the tape is conceptually infinite)
		while(*lRightCounter == *rRightCounter && lRightCounter != tape.end() && rRightCounter != rhs.tape.end())
		{
			++lRightCounter;
			++rRightCounter;
		}
		if(lRightCounter == tape.end() && rRightCounter == rhs.tape.end())
			go = true;
		else if(lRightCounter == tape.end())
		{
			while(*rRightCounter == rhs.default_value && rRightCounter != rhs.tape.end())
				++rRightCounter;
			if(rRightCounter == rhs.tape.end())
				go = true;
		}
		else if(rRightCounter == rhs.tape.end())
		{
			while(*lRightCounter == default_value && lRightCounter != tape.end())
				++lRightCounter;
			if(lRightCounter == tape.end())
				go = true;
		}
		else
			go = false;
		//--//

		//if above check fails, return false
		if(!go)
			return false;


		//--check to the left of current
		//the following configuration is required: C, X, Y, Z are words, A is one symbol -><- represents read head
		//this: 0*CX->A<-Y0*
		//rhs:  0*CZX->A<-Y0*
		//one additional requirement is that lhsLeftBound must be to the right of C in this
		while(*lLeftCounter == *rLeftCounter && lLeftCounter != tape.begin() && rLeftCounter != rhs.tape.begin())
		{
			--lLeftCounter;
			--rLeftCounter;
		}
		if(lLeftCounter == tape.begin() && rLeftCounter == rhs.tape.begin())
			go = true;
		else if(lLeftCounter == tape.begin())
			go = true;
		else if(rLeftCounter == rhs.tape.begin())
		{
			while(*lLeftCounter == default_value && lLeftCounter != tape.begin())
				--lLeftCounter;
			if(lLeftCounter == tape.begin())
				go = true;
		}
		//else if(*lLeftCounter != default_value)
		//	go = false;
		else
		{
			typename std::list<T>::iterator lTempCounter = tape.begin();
			typename std::list<T>::iterator rTempCounter = rhs.tape.begin();
			while(*(++lTempCounter) == default_value && lTempCounter != lLeftCounter)
			{
				if(lTempCounter == lhsLeftBound)
					return false;
			}
			while(*(++rTempCounter) == rhs.default_value && rTempCounter != rLeftCounter);
			while(lTempCounter != lLeftCounter && rTempCounter != rLeftCounter)
			{
				if(lTempCounter == lhsLeftBound || *lTempCounter != *rTempCounter || rTempCounter == rLeftCounter)
					return false;
				lTempCounter++;
				rTempCounter++;
			}
			if(lLeftCounter == lhsLeftBound || rTempCounter == rLeftCounter)
				return false;
			go = true;
		}
		//--//
		return go;
	}

	/**
	 * tests for a simple loop repeat configuration in rhs (mirror of repeatConfigRight).
	 * @pre this and rhs represent two points during a particular turing_machine execution; the current state at each point in the execution is the same; the current read head symbol at each point in the execution is the same;
	 * @post true if this tape and rhs having matching configurations that exhibit left motion simple loop behavior
	 * @param rhs representation of the tape at some point later on in the execution
	 * @param lhsRightBound farthest cell to the right on the tape that the read head has been between the two points of execution
	 * @author Owen Kellett
	 */
	bool repeatConfigLeft(infinite_tape<T> &rhs, typename std::list<T>::iterator lhsRightBound)
	{
		typename std::list<T>::iterator lRightCounter = current;
		typename std::list<T>::iterator lLeftCounter = current;
		typename std::list<T>::iterator rRightCounter = rhs.current;
		typename std::list<T>::iterator rLeftCounter = rhs.current;
		bool go = false;

		//check to the left of current
		while(*lLeftCounter == *rLeftCounter && lLeftCounter != tape.begin() && rLeftCounter != rhs.tape.begin())
		{
			--lLeftCounter;
			--rLeftCounter;
		}
		if(lLeftCounter == tape.begin() && rLeftCounter == rhs.tape.begin())
			go = true;
		else if(lLeftCounter == tape.begin())
		{
			while(*rLeftCounter == rhs.default_value && rLeftCounter != rhs.tape.begin())
				--rLeftCounter;
			if(rLeftCounter == rhs.tape.begin())
				go = true;
		}
		else if(rLeftCounter == rhs.tape.begin())
		{
			while(*lLeftCounter == default_value && lLeftCounter != tape.begin())
				--lLeftCounter;
			if(lLeftCounter == tape.begin())
				go = true;
		}
		else
			go = false;

		if(!go)
			return false;

		//check to the right of current
		while(*lRightCounter == *rRightCounter && lRightCounter != tape.end() && rRightCounter != rhs.tape.end())
		{
			++lRightCounter;
			++rRightCounter;
		}
		if(lRightCounter == tape.end() && rRightCounter == rhs.tape.end())
			go = true;
		else if(lRightCounter == tape.end())
			go = true;
		else if(rRightCounter == rhs.tape.end())
		{
			while(*lRightCounter == default_value && lRightCounter != tape.end())
				++lRightCounter;
			if(lRightCounter == tape.end())
				go = true;
		}
		//else if(*lRightCounter != default_value)
			//go = false;
		else
		{
			typename std::list<T>::iterator lTempCounter = tape.end();
			typename std::list<T>::iterator rTempCounter = rhs.tape.end();
			while(*(--lTempCounter) == default_value && lTempCounter != lRightCounter)
			{
				//std::cout << *lTempCounter << " " << *lhsRightBound << std::endl;
				if(lTempCounter == lhsRightBound)
					return false;
			}
			while(*(--rTempCounter) == rhs.default_value && rTempCounter != rRightCounter);
			while(lTempCounter != lRightCounter && rTempCounter != rRightCounter)
			{
				//std::cout << *lTempCounter << " " << *lhsRightBound << std::endl;
				if(lTempCounter == lhsRightBound || *lTempCounter != *rTempCounter || rTempCounter == rRightCounter)
					return false;
				lTempCounter--;
				rTempCounter--;
			}
			//std::cout << *lTempCounter << " " << *lhsRightBound << std::endl;
			if(lRightCounter == lhsRightBound || rTempCounter == rRightCounter)
				return false;
			go = true;
		}
		return go;
	}

	/**
	 * Expands the tape if the iterator is at the left edge of the tape.
	 * @param leftEdge location on the tape to check
	 * @author Owen Kellett
	 */
	void checkIteratorLeft(typename std::list<T>::iterator leftEdge)
	{
		typename std::list<T>::iterator tempIter(leftEdge);
		if((--tempIter) == tape.begin())
		{
			expand_tape(left, 10);
			//cout << "expanded tape " << counter << " to the left" << endl;
		}
	}

	/**
	 * Expands the tape if the iterator is at the right edge of the tape.
	 * @param rightEdge location on the tape to check
	 * @author Owen Kellett
	 */
	void checkIteratorRight(typename std::list<T>::iterator rightEdge)
	{
		typename std::list<T>::iterator tempIter(rightEdge);
		if((++tempIter) == tape.end())
		{
			expand_tape(right, 10);
			//cout << "expanded tape " << counter << " to the right" << endl;
		}
	}

	//end Owen Kellett modifications

protected:
	//private functions used to maintain invariants regarding the tape to ensure correct operation

	/**
	 * check to see if tape needs to be expanded and call expansion function if so.
	 * @pre current refers to a valid cell on the tape, which==left xor which==right
	 * @post if which==left then it is safe to move left at least one position on the tape; if which==right then it is safe to move at least one position right on the tape
	 * @param which direction to check
	 * @author Kyle Ross
	 */
	void tape_check(direction which)
	{
		if(which==left)
		{
			typename std::list<T>::reverse_iterator temp_iter(current);
			if((++temp_iter)==tape.rend())
			{
				#ifdef ross_debug_mode
					std::cout<<"["<<this<<"] expanding on left"<<std::endl;
				#endif
				if(counter<1) //should never happen if tape is created with at least three cells ... but just to be sure
				{
					counter=1;
				}
				expand_tape(left, counter);
			}
		}
		else
		{
			typename std::list<T>::iterator temp_iter(current);
			if((++temp_iter)==tape.end())
			{
				#ifdef ross_debug_mode
					std::cout<<"["<<this<<"] expanding on right"<<std::endl;
				#endif
				if(counter<1)
				{
					counter=1;
				}
				expand_tape(right, counter);
			}
			else
			{
				#ifdef ross_debug_mode
					#ifdef ross_debug_verbose
						std::cout<<"["<<this<<"]no expansion needed"<<std::endl;
					#endif
				#endif
				return;
			}
		}
	}

	/**
	 * Expands the tape.
	 * @post the tape has been expanded by how_many to the left if which==left xor has been expanded how_many to the right if which==right
	 * @param which direction to expand
	 * @param how_many number of additional cells to add
	 * @author Kyle Ross
	 */
	void expand_tape(direction which, typename std::list<T>::size_type how_many)
	{
		if(which==left)
		{
			tape.insert(tape.begin(), how_many, default_value);
		}
		else if(which==right)
		{
			tape.insert(tape.end(), how_many, default_value);
		}
		#ifdef ross_debug_mode
			std::cout<<"["<<this<<"]expanded by "<<counter<<" on "<<(which==left?"left":"right")<<" to "<<tape.size()<<std::endl;
		#endif
		counter=0;
	}

	/**
	 * arbitrary start size of tape.
	 * odd number chosen so that there are 50 cells to each of the right and left of the cardinal ... just to make the numbers a bit easier
	 */
	static const typename std::list<T>::size_type start_size=101;
	/**
	 * the cells of the tape
	 */
	typename std::list<T> tape;
	/**
	 * location of read-head on tape
	 */
	typename std::list<T>::iterator current, cardinal;
	/**
	 * number of moves made counter.
	 * allows constant amortised time moves with expansion
	 */
	typename std::list<T>::size_type counter;
	/**
	 * default value when creating a new cell
	 */
	T default_value;
	/**
	 * count the number of non-blank symbols
	 */
	typename std::list<T>::size_type non_blank_count;
};

/**
 * function for use in displaying an infinite_tape object.
 * @pre only those enforced by the show member function of infinite_tape
 * @post those of the show member function of infinite_tape; out is returned
 * @param out output stream to print to
 * @param to_print tape to print
 * @author Kyle Ross
 * @ingroup TuringMachineDefinition
 */
template <typename T>
std::ostream& operator<<(std::ostream &out, const infinite_tape<T> &to_print)
{
	to_print.show(out);
	return out;
}

#endif

