#ifndef ross_turing_machine_header
#define ross_turing_machine_header

/*Authored 2002    by  Kyle  Ross,  Undergraduate  Philosophy   Thesis
(Advisor: Professor Bram van Heuveln)

Modified 2003    by Owen Kellett, Undergraduate Research
(Advisor: Professor Bram van Heuveln)

This  file is part of the   "Generic Turing Machine Implementation for
Use in the Exploration of the Busy Beaver Problem".

The "Generic Turing Machine Implementation for  Use in the Exploration
of the Busy Beaver Problem" is free  software; you can redistribute it
and/or modify it under the terms of the  GNU General Public License as
published   by the Free Software Foundation;   either version 2 of the
License, or (at your option) any later version.

This programme is distributed in the hope that it  will be useful, but
WITHOUT    ANY  WARRANTY;  without   even  the    implied warranty  of
MERCHANTABILITY or FITNESS FOR  A  PARTICULAR  PURPOSE.  See  the  GNU
General Public License for more details.

You should have received   a copy of  the  GNU General Public  License
along  with  this   programme; if  not,  write   to the Free  Software
Foundation, Inc., 59  Temple Place, Suite  330, Boston, MA  02111-1307
USA (or visit http://www.fsf.org/licenses/licenses.html)

Created as part of the Busy Beaver Project:
Bram van Heuveln <heuveb@rpi.edu>
Boleshaw Szymanski <szymansk@cs.rpi.edu>
Selmer Bringsjord <selmer@rpi.edu>
Carlos Varela <cvarela@cs.rpi.edu>
Owen Kellett <kelleo@rpi.edu>
Shailesh Kelkar <kelkas@rpi.edu>
Kyle Ross <rossk2@rpi.edu>
Rensselaer Reasoning Group
Department of Cognitive Science
Department of Computer Science
Rensselaer Polytechnic Institute
Troy, New York 12180
U.S.A.
*/
//implements a five-tuple turing machine (and can simulate a four-tuple turing machine) over an arbitrary (small) alphabet

#define enableLoggingOutput
#define enable_optimisations
#define blank_tape_optimisation
#define no_ss_optimisation
#define no_ssprime_sprimes_optimisation
#define first_write_optimisation
#define first_move_optimisation

void display_optimisation_settings(std::ostream &out)
{
	#ifdef enable_optimisations
		out<<"optimisations enabled:"<<std::endl;
		out<<" blank-tape-optimisation: ";
		#ifdef blank_tape_optimisation
			out<<"on"<<std::endl;
		#else
			out<<"off"<<std::endl;
		#endif
		out<<" no-ss-optimisation: ";
		#ifdef no_ss_optimisation
			out<<"on"<<std::endl;
		#else
			out<<"off"<<std::endl;
		#endif
		out<<" no-ssprime-sprimes-optimisation: ";
		#ifdef no_ssprime_sprimes_optimisation
			out<<"on"<<std::endl;
		#else
			out<<"off"<<std::endl;
		#endif
		out<<" first-write-optimisation: ";
		#ifdef first_write_optimisation
			out<<"on"<<std::endl;
		#else
			out<<"off"<<std::endl;
		#endif
		out<<" first-move-optimisation: ";
		#ifdef first_move_optimisation
			#ifndef first_write_optimisation
				out<<"disabled because first-write is not on"<<std::endl;
			#else
				out<<"on"<<std::endl;
			#endif
		#else
			out<<"off"<<std::endl;
		#endif
	#else
		out<<"optimisations disabled"<<std::endl;
	#endif
}

#include <cmath>
#include <vector>
#include "debug.hpp"
#include "infinite_tape.hpp"
#include "state.hpp"
#include "machine_storage.hpp"
#include "watch.hpp"
template <typename T>
T& last_element(std::vector<T> &the_list);

/**
 * turing machine with an arbitrary (but small) alphabet.
 * @ingroup TuringMachineDefinition
 */
template <typename T>
class turing_machine
{
public:
	typedef typename std::vector<state<T> >::size_type transition_index_type;
	typedef typename infinite_tape<T>::direction move_type;
	typedef typename state<T>::transition_result result_type;
	typedef unsigned long int transition_count_type;
	static const move_type left=infinite_tape<T>::left;
	static const move_type right=infinite_tape<T>::right;
	static const move_type do_nothing=infinite_tape<T>::do_nothing;
	static const move_type halt_move=infinite_tape<T>::halt_move;
	static const result_type run=state<T>::run;
	static const result_type halt=state<T>::halt;

	/**
	 * copies a turing machine.
	 * @pre rhs is a valid turing_machine
	 * @post *this is a functionally-equivalent turing_machine to rhs
	 * @param rhs machine to be copied
	 * @author Kyle Ross
	 */
	turing_machine(const turing_machine<T> &rhs)
	:current(rhs.current), transition_function(rhs.transition_function), tape(rhs.tape), steps(rhs.steps), run_status(rhs.run_status), halt_state(rhs.halt_state), halt_state_exists(rhs.halt_state_exists), first_unreachable_state(rhs.first_unreachable_state)
	{
		#ifdef ross_debug_mode
			std::cout<<"["<<this<<"]turing-machine copy-constructed"<<std::endl;
		#endif
	}

	/**
	 * default-most constructor.
	 * creates a turing machine from T representing default tape value
	 * @post *this refers to a turing_machine with a default-constructed tape that has default_tape_value as its blank-symbol
	 * @param default_tape_value default value to be set on the tape
	 * @author Kyle Ross
	 */
	turing_machine(T default_tape_value)
	:tape(default_tape_value), current(0), steps(0), run_status(run), halt_state_exists(false), first_unreachable_state(1)
	{
		#ifdef ross_debug_mode
			std::cout<<"["<<this<<"]turing-machine constructed"<<std::endl;
		#endif
	}

	/**
	 * creates an N-state stturing_machine whose tape has a given T as its default value
	 * @pre N>=0
	 * @post a turing machine with N default states has been constructed with a default-size tape whose blank-symbol is default_tape_value
	 * @param N number of states
	 * @param default_tape_value default value to be set on the tape
	 * @author Kyle Ross
	 */
	turing_machine(transition_index_type N, T default_tape_value)
	:tape(default_tape_value), current(0), transition_function(N), steps(0), run_status(run), halt_state_exists(false), first_unreachable_state(1)
	{
		#ifdef ross_debug_mode
			std::cout<<"["<<this<<"]turing-machine constructed"<<std::endl;
		#endif
	}

	/**
	 * creates a machine with an empty set of states and the given tape.
	 * @pre new_tape is a valid infinite_tape
	 * @post a turing_machine with no states has been constructed with a tape that is functionally-equivalent to new_tape
	 * @param new_tape tape to be used for the new machine
	 * @author Kyle Ross
	 */
	turing_machine(const infinite_tape<T> &new_tape)
	:tape(new_tape), current(0), steps(0), run_status(run), halt_state_exists(false), first_unreachable_state(1)
	{
		#ifdef ross_debug_mode
			std::cout<<"["<<this<<"]turing-machine constructed"<<std::endl;
		#endif
	}

	/**
	 * trivial destructor
	 * @author Kyle Ross
	 */
	~turing_machine()
	{
		#ifdef ross_debug_mode
			std::cout<<"["<<this<<"]turing-machine destroyed"<<std::endl;
		#endif
	}

	/**
	 * adds an additional state to the transition function.
	 * @pre there are zero or more states in the FA of the turing_machine
	 * @post an additional default state has been added to the turing_machine's FA and the current number of states in the FA is returned
	 * @return the current number of states in the machine
	 * @author Kyle Ross
	 */
	transition_index_type add_state()
	{
		transition_function.push_back(state<T>()); //calling the default constructor to create an anonymous object each time might be more overhead than having a static default state<T>() stored, but I'm not going to worry about this since add_state calls' cost should be trivial compared to running the TM
		return transition_function.size();
	}

	/**
	 * set the halt state.
	 * @param new_halt_state index of the state to set as halt
	 * @return true if halt state successfully set
	 * @author Kyle Ross
	 */
	bool set_halt_state(transition_index_type new_halt_state)
	{
		if(halt_state_exists)
		{
			#ifdef ross_debug_mode
				std::cout<<"["<<this<<"] could not set halt state since one already exists"<<std::endl;
			#endif
			return false;
		}
		else
		{
			if(bounds_check(new_halt_state))
			{
				halt_state_exists=true;
				halt_state=new_halt_state;
				#ifdef ross_debug_mode
					std::cout<<"["<<this<<"] halt state set successfully"<<std::endl;
				#endif
				return true;
			}
			else
			{
				#ifdef ross_debug_mode
					std::cout<<"["<<this<<"] could not create halt state since index was invalid"<<std::endl;
				#endif
				return false;
			}
		}
	}

	/**
	 * adds a full transition.
	 * @post a new full transition as defined by the parameter values has been created iff each of start_state, end_state are >=0 and <transition_function.size()
	 * @param start_state starting state of transition
	 * @param read read symbol for transition
	 * @param write symbol to be written on the tape
	 * @param move direction to move read head
	 * @end_state new state to transition to
	 * @return true if transition successfully added
	 * @author Kyle Ross
	 */
	bool add_transition(transition_index_type start_state, T read, T write, move_type move, transition_index_type end_state)
	{
		if(bounds_check(start_state) && bounds_check(end_state)) //are both states valid?
		{
			#ifdef ross_debug_mode
				std::cout<<"["<<this<<"]turing-machine::add-transition to state "<<start_state<<std::endl;
			#endif
			transition_function[start_state].create_transition(read, write, end_state, move);
			if(end_state!=halt_state && end_state>=first_unreachable_state)
			{
				first_unreachable_state=end_state+1;
			}
			return true;
		}
		else
		{
			#ifdef ross_debug_mode
				std::cout<<"["<<this<<"]turing-machine::add-transition WARNING transition not added because of bounds-check failure"<<std::endl;
			#endif
			return false;
		}
	}

	/**
	 * adds a 4-tuple write-only transition
	 * @post a new full transition as defined by the parameter values has been created iff each of start_state, end_state are >=0 and <transition_function.size()
	 * @param start_state starting state of transition
	 * @param read read symbol for transition
	 * @param write symbol to be written on the tape
	 * @end_state new state to transition to
	 * @return true if transition successfully added
	 * @author Kyle Ross
	 */
	bool add_write_transition(transition_index_type start_state, T read, T write, transition_index_type end_state)
	{
		return add_transition(start_state, read, write, do_nothing, end_state);
	}

	/**
	 * adds a 4-tuple move-only transition
	 * @post a new full transition as defined by the parameter values has been created iff each of start_state, end_state are >=0 and <transition_function.size()
	 * @param start_state starting state of transition
	 * @param read read symbol for transition
	 * @param move direction to move read head
	 * @end_state new state to transition to
	 * @return true if transition successfully added
	 * @author Kyle Ross
	 */
	bool add_move_transition(transition_index_type start_state, T read, move_type move, transition_index_type end_state)
	{
		return add_transition(start_state, read, read, move, end_state);
	}

	/**
	 * returns whether to_check refers to an existing state in the FA or not
	 * @pre transition_function holds zero or more states
	 * @post true has been returned iff 0<=to_check<transition_function.size(), false else
	 * @param to_check state to check
	 * @author Kyle Ross
	 */
	bool bounds_check(transition_index_type to_check)
	{
		return to_check<transition_function.size();
	}

	/**
	 * run one step and return whether the machine is running or has halted
	 * @pre bounds_check(current)==true, the tape is in a valid state
	 * @post the transition, given the current read-symbol, is processed properly and the current run-status of the machine is returned
	 * @return whether or not the machine is running
	 * @author Kyle Ross
	 */
	result_type run_one()
	{
		if((run_status=transition_function[current].perform_transition(tape, current))==run)
		{
			++steps;
		}
		if(halt_state_exists && current==halt_state)
		{
			run_status=halt;
		}
		return run_status;
	}

	//functionality for accessing the infinite_tape

	/**
	 * count occurrences of to_count on the tape ... e.g. to find productivity.
	 * @pre the tape has zero or more cells
	 * @post the number of occurrences of to_count on the tape have been returned
	 * @param to_count symbol to count
	 * @return number of occurences of to_count on the tape
	 * @author Kyle Ross
	 */
	transition_index_type count_occurrences(T to_count) const
	{
		return tape.count_occurrences(to_count);
	}

	/**
	 * provides fairly low-overhead read-only access to the infinite_tape.
	 * @return constant reference to the tape
	 * @author Kyle Ross
	 */
	const infinite_tape<T>& access_tape() const
	{
		return tape;
	}

	/**
	 * provides read-write access to the tape.
	 * hmmm, I can't really justify this function except that it's sometimes useful to mangle the contents of the tape ... it's a gross violation of encapsulation, so it should be used *very* carefully
	 * used carefully by Owen in non-halt routines :)
	 * @return reference to the tape
	 * @author Kyle Ross
	 */
	infinite_tape<T>& access_tape_read_write()
	{
		#ifdef ross_debug_mode
			std::cout<<"["<<this<<"]turing-machine::access-tape-read-write WARNING this function should be used carefully"<<std::endl;
		#endif
		return tape;
	}

	/**
	 * determine if the tape is in standard position or not.
	 * standard position is defined as: the read head is on either the leftmost or rightmost symbol ... may also be accomplished (by user) in the following way: turing_machine<T> x; x.access_tape.in_standard_position();
	 * @return true is returned iff the tape is in standard position as defined in the infinite_tape.in_standard_position
	 * @author Kyle Ross
	 */
	bool in_standard_position() const
	{
		return tape.in_standard_position();
	}

	/**
	 * determine if the tape is blank or not.
	 * @return true is returned iff the tape is blank; false else
	 * @author Kyle Ross
	 */
	bool blank_tape() const
	{
		return tape.is_blank();
	}

	/**
	 * provide access to certain information about the FA.
	 * @return how many states the FA has
	 * @author Kyle Ross
	 */
	transition_index_type count_states() const
	{
		return transition_function.size();
	}

	/**
	 * returns the number of the FA's current state
	 * @return the number of the FA's current state
	 * @author Kyle Ross
	 */
	transition_index_type get_current_state() const
	{
		return current;
	}

	/**
	 * whether or not the machine is running.
	 * @post the run status of the machine has been returned, defined to be run if no transitions have been executed
	 * @return run status of the machine
	 * @author Kyle Ross
	 */
	result_type get_status() const
	{
		return run_status;
	}

	/**
	 * returns how many times the transition function has been called.
	 * @return number of transitions the machine has made
	 * @author Kyle Ross
	 */
	transition_count_type get_steps_count() const
	{
		return steps;
	}

	/**
	 * returns if the machine is in the halt state
	 * @return true if the machine is in the halt state
	 * @author Kyle Ross
	 */
	bool in_halt_state() const
	{
		return halt_state_exists && current==halt_state;
	}

	/**
	 * number of equivalent machines that this machine represents
	 * @return number of equivalent machines that this machine represents
	 * @author Kyle Ross
	 */
	mpz_class number_represented() const
	{
		unsigned int m=0;
		unsigned int n=count_states()-1;

		for(typename std::vector<state<T> >::const_iterator i=transition_function.begin(); (i+1)!=transition_function.end(); ++i)
		{
			for(T j=0; j<=1; ++j)
			{
				T write;
				typename infinite_tape<T>::direction move;
				typename std::vector<state<T> >::size_type index;
				if((*i).simulate_transition(j, write, move, index)==run)
				{
					++m;
				}
			}
		}

		return permutations(n-1, first_unreachable_state-1)*pow(mpz_class(4*n+1), mpz_class(2*n-m));
	}

	/**
	 * number of equivalent machines that this machine represents
	 * @return number of equivalent machines that this machine represents in the explicit formulation
	 * @author Owen Kellett
	 */
	mpz_class number_represented_explicit() const
	{
		unsigned int m=0;
		unsigned int n=count_states()-1;

		for(typename std::vector<state<T> >::const_iterator i=transition_function.begin(); (i+1)!=transition_function.end(); ++i)
		{
			for(T j=0; j<=1; ++j)
			{
				T write;
				typename infinite_tape<T>::direction move;
				typename std::vector<state<T> >::size_type index;
				if((*i).simulate_transition(j, write, move, index)==run)
				{
					++m;
				}
			}
		}

		return permutations(n-1, first_unreachable_state-1)*pow(mpz_class(4*n+4), mpz_class(2*n-m));
	}



	//functions for output of the turing machine in various formats

	/**
	 * prints the turing machine in a rather messy (but correct and informative) form.
	 * @author Kyle Ross
	 */
	void print(std::ostream &out) const
	{
		out<<tape<<std::endl;
		typename std::vector<state<T> >::const_iterator i=transition_function.begin();
		if(i==transition_function.end())
		{
			out<<"< no states exist >"<<std::endl;
		}
		else
		{
			for(int count(0); i!=transition_function.end(); ++i)
			{
				out<<"state "<<(count++)<<std::endl<<(*i);
			}
			out<<"current state is "<<current<<std::endl;
			if(halt_state_exists)
			{
				out<<"halt state is "<<halt_state<<std::endl;
			}
			out<<"first unreachable state is "<<first_unreachable_state<<std::endl;
		}
		out<<"current run status is: "<<(run_status==run?"running":"halted")<<" after "<<steps<<" transitions"<<std::endl;
	}
	//functionality for tree normalisation method
	/**
	 * is it possible to tree-normalise this machine?
	 * @return true if machine is halted in a non-halting state
	 * @author Kyle Ross
	 */
	bool is_tn_possible() const
	{
		if(halt_state_exists)
		{
			return run_status==halt && current!=halt_state;
		}
		else
		{
			return run_status==halt;
		}
	}

	/**
	 * creates children in the tree for quadruple implicit halt
	 * @post quadruple implicit child machines for this machine have been outputed to destination
	 * @param alphabet alphabet for which the machines are defined
	 * @param destination mechanism to load child machines onto stack
	 * @author Kyle Ross
	 */
	template <typename OutputIterator>
	void generate_tn_quadruple_children(const std::list<T> &alphabet, OutputIterator destination) const
	{
		//store read symbol
		T tape_symbol=tape.read();

		//create a temp machine for use in generating children
		turing_machine<T> temp_machine=*this;

		//for each reachable (used) state:
		for(transition_index_type end_state=0; end_state!=first_unreachable_state; ++end_state)
		{
			//for each possible write & move, create a child machine
			for(typename std::list<T>::const_iterator alpha_iter=alphabet.begin(); alpha_iter!=alphabet.end(); ++alpha_iter)
			{
				#ifdef no_ss_optimisation
					if((*alpha_iter)!=tape_symbol)
					{
						temp_machine=*this;
						temp_machine.add_write_transition(current, tape_symbol, *alpha_iter, end_state);
						output_machine(temp_machine, destination, alphabet);
					}
					else
					{
						temp_machine=*this;
						temp_machine.add_write_transition(current, tape_symbol, *alpha_iter, end_state);
						g_counter_implicit.increment(tm_stats::simple_nonproductive, temp_machine.number_represented());
						c_counter_implicit.increment(tm_stats::simple_nonproductive, 1);
					}
				#else
						temp_machine=*this;
						temp_machine.add_write_transition(current, tape_symbol, *alpha_iter, end_state);
						output_machine(temp_machine, destination, alphabet);
				#endif
			}

			//for each move, create a child machine
			temp_machine=*this;
			temp_machine.add_move_transition(current, tape_symbol, left, end_state);
			output_machine(temp_machine, destination, alphabet);
			temp_machine=*this;
			temp_machine.add_move_transition(current, tape_symbol, right, end_state);
			output_machine(temp_machine, destination, alphabet);
		}

		//if this machine has a halt state (which, presumably, it should!)
		if(halt_state_exists)
		{
			//if there is an unreachable state, create transitions to it
			if(first_unreachable_state!=halt_state)
			{
				for(typename std::list<T>::const_iterator alpha_iter=alphabet.begin(); alpha_iter!=alphabet.end(); ++alpha_iter)
				{
					#ifdef no_ss_optimisation
						if((*alpha_iter)!=tape_symbol)
						{
							temp_machine=*this;
							temp_machine.add_write_transition(current, tape_symbol, *alpha_iter, first_unreachable_state);
							output_machine(temp_machine, destination, alphabet);
						}
						else
						{
							temp_machine=*this;
							temp_machine.add_write_transition(current, tape_symbol, *alpha_iter, first_unreachable_state);
							g_counter_implicit.increment(tm_stats::simple_nonproductive, temp_machine.number_represented());
							c_counter_implicit.increment(tm_stats::simple_nonproductive, 1);
						}
					#else
						temp_machine=*this;
						temp_machine.add_write_transition(current, tape_symbol, *alpha_iter, first_unreachable_state);
						output_machine(temp_machine, destination, alphabet);
					#endif
				}
				temp_machine=*this;
				temp_machine.add_move_transition(current, tape_symbol, left, first_unreachable_state);
				output_machine(temp_machine, destination, alphabet);
				temp_machine=*this;
				temp_machine.add_move_transition(current, tape_symbol, right, first_unreachable_state);
				output_machine(temp_machine, destination, alphabet);
			}

			//create transition to the halt state
			temp_machine=*this;
			temp_machine.add_write_transition(current, tape_symbol, tape_symbol, halt_state);
			output_machine(temp_machine, destination, alphabet);
		}
	}

	/**
	 * creates children in the tree for quadruple explicit halt
	 * @post quadruple explicit child machines for this machine have been outputed to destination
	 * @param alphabet alphabet for which the machines are defined
	 * @param destination mechanism to load child machines onto stack
	 * @author Kyle Ross
	 */
	template <typename OutputIterator>
	void generate_tn_quadruple_children_explicit(const std::list<T> &alphabet, OutputIterator destination) const
	{
		//store read symbol
		T tape_symbol=tape.read();

		//create a temp machine for use in generating children
		turing_machine<T> temp_machine=*this;

		//for each reachable (used) state:
		for(transition_index_type end_state=0; end_state!=first_unreachable_state; ++end_state)
		{
			//for each possible write & move, create a child machine
			for(typename std::list<T>::const_iterator alpha_iter=alphabet.begin(); alpha_iter!=alphabet.end(); ++alpha_iter)
			{
				#ifdef no_ss_optimisation
					if((*alpha_iter)!=tape_symbol)
					{
						temp_machine=*this;
						temp_machine.add_write_transition(current, tape_symbol, *alpha_iter, end_state);
						output_machine(temp_machine, destination, alphabet);
					}
				#else
					temp_machine=*this;
					temp_machine.add_write_transition(current, tape_symbol, *alpha_iter, end_state);
					output_machine(temp_machine, destination, alphabet);
				#endif
			}

			//for each move, create a child machine
			temp_machine=*this;
			temp_machine.add_move_transition(current, tape_symbol, left, end_state);
			output_machine(temp_machine, destination, alphabet);
			temp_machine=*this;
			temp_machine.add_move_transition(current, tape_symbol, right, end_state);
			output_machine(temp_machine, destination, alphabet);
		}

		//if this machine has a halt state (which, presumably, it should!)
		if(halt_state_exists)
		{
			//if there is an unreachable state, create transitions to it
			if(first_unreachable_state<halt_state)
			{
				for(typename std::list<T>::const_iterator alpha_iter=alphabet.begin(); alpha_iter!=alphabet.end(); ++alpha_iter)
				{
					#ifdef no_ss_optimisation
						if((*alpha_iter)!=tape_symbol)
						{
							temp_machine=*this;
							temp_machine.add_write_transition(current, tape_symbol, *alpha_iter, first_unreachable_state);
							output_machine(temp_machine, destination, alphabet);
						}
					#else
						temp_machine=*this;
						temp_machine.add_write_transition(current, tape_symbol, *alpha_iter, first_unreachable_state);
						output_machine(temp_machine, destination, alphabet);
					#endif
				}
				temp_machine=*this;
				temp_machine.add_move_transition(current, tape_symbol, left, first_unreachable_state);
				output_machine(temp_machine, destination, alphabet);
				temp_machine=*this;
				temp_machine.add_move_transition(current, tape_symbol, right, first_unreachable_state);
				output_machine(temp_machine, destination, alphabet);
			}

			//create transitions to the halt state
			/*temp_machine=*this;
			temp_machine.add_write_transition(current, tape_symbol, tape_symbol, halt_state);
			output_machine(temp_machine, destination, alphabet);*/
			for(typename std::list<T>::const_iterator alpha_iter=alphabet.begin() ; alpha_iter!=alphabet.end(); ++alpha_iter)
			{
				#ifdef no_ss_optimisation
				 if((*alpha_iter)!=tape_symbol)
					{
						temp_machine=*this;
						temp_machine.add_write_transition(current, tape_symbol, *alpha_iter, halt_state);
						output_machine(temp_machine, destination, alphabet);
					}
				#else
					temp_machine=*this;
					temp_machine.add_write_transition(current, tape_symbol, *alpha_iter, halt_state);
					output_machine(temp_machine, destination, alphabet);
				#endif
			}
			temp_machine=*this;
			temp_machine.add_move_transition(current, tape_symbol, left, halt_state);
			output_machine(temp_machine, destination, alphabet);
			temp_machine=*this;
			temp_machine.add_move_transition(current, tape_symbol, right, halt_state);
			output_machine(temp_machine, destination, alphabet);
		}
	}

	/**
	 * creates all children (implicit and explicit) in the tree for the quadruple formulation.
	 * @post quadruple child machines for this machine have been outputed to destination
	 * @param alphabet alphabet for which the machines are defined
	 * @param destination mechanism to load child machines onto stack
	 * @author Owen Kellett
	 */
	template <typename OutputIterator>
	void generate_all_tn_quadruple_children(const std::list<T> &alphabet, OutputIterator destination) const
	{
		//store read symbol
		T tape_symbol=tape.read();

		//create a temp machine for use in generating children
		turing_machine<T> temp_machine=*this;

		//for each reachable (used) state:
		for(transition_index_type end_state=0; end_state!=first_unreachable_state; ++end_state)
		{
			//for each possible write & move, create a child machine
			for(typename std::list<T>::const_iterator alpha_iter=alphabet.begin(); alpha_iter!=alphabet.end(); ++alpha_iter)
			{
				#ifdef no_ss_optimisation
					if((*alpha_iter)!=tape_symbol)
					{
						temp_machine=*this;
						temp_machine.add_write_transition(current, tape_symbol, *alpha_iter, end_state);
						output_machine(temp_machine, destination, alphabet);
					}
					else
					{
						temp_machine=*this;
						temp_machine.add_write_transition(current, tape_symbol, *alpha_iter, end_state);
						g_counter_implicit.increment(tm_stats::simple_nonproductive, temp_machine.number_represented());
						c_counter_implicit.increment(tm_stats::simple_nonproductive, 1);
						g_counter_explicit.increment(tm_stats::simple_nonproductive, temp_machine.number_represented_explicit());
						c_counter_explicit.increment(tm_stats::simple_nonproductive, 1);
					}
				#else
						temp_machine=*this;
						temp_machine.add_write_transition(current, tape_symbol, *alpha_iter, end_state);
						output_machine(temp_machine, destination, alphabet);
				#endif
			}

			//for each move, create a child machine
			temp_machine=*this;
			temp_machine.add_move_transition(current, tape_symbol, left, end_state);
			output_machine(temp_machine, destination, alphabet);
			temp_machine=*this;
			temp_machine.add_move_transition(current, tape_symbol, right, end_state);
			output_machine(temp_machine, destination, alphabet);
		}

		//if this machine has a halt state (which, presumably, it should!)
		if(halt_state_exists)
		{
			//if there is an unreachable state, create transitions to it
			if(first_unreachable_state!=halt_state)
			{
				for(typename std::list<T>::const_iterator alpha_iter=alphabet.begin(); alpha_iter!=alphabet.end(); ++alpha_iter)
				{
					#ifdef no_ss_optimisation
						if((*alpha_iter)!=tape_symbol)
						{
							temp_machine=*this;
							temp_machine.add_write_transition(current, tape_symbol, *alpha_iter, first_unreachable_state);
							output_machine(temp_machine, destination, alphabet);
						}
						else
						{
							temp_machine=*this;
							temp_machine.add_write_transition(current, tape_symbol, *alpha_iter, first_unreachable_state);
							g_counter_implicit.increment(tm_stats::simple_nonproductive, temp_machine.number_represented());
							c_counter_implicit.increment(tm_stats::simple_nonproductive, 1);
							g_counter_explicit.increment(tm_stats::simple_nonproductive, temp_machine.number_represented_explicit());
							c_counter_explicit.increment(tm_stats::simple_nonproductive, 1);
						}
					#else
						temp_machine=*this;
						temp_machine.add_write_transition(current, tape_symbol, *alpha_iter, first_unreachable_state);
						output_machine(temp_machine, destination, alphabet);
					#endif
				}
				temp_machine=*this;
				temp_machine.add_move_transition(current, tape_symbol, left, first_unreachable_state);
				output_machine(temp_machine, destination, alphabet);
				temp_machine=*this;
				temp_machine.add_move_transition(current, tape_symbol, right, first_unreachable_state);
				output_machine(temp_machine, destination, alphabet);
			}

			//create transition to the halt state
			for(typename std::list<T>::const_iterator alpha_iter=alphabet.begin() ; alpha_iter!=alphabet.end(); ++alpha_iter)
			{
				temp_machine=*this;
				temp_machine.add_write_transition(current, tape_symbol, *alpha_iter, halt_state);
				output_machine(temp_machine, destination, alphabet);
			}
			temp_machine=*this;
			temp_machine.add_move_transition(current, tape_symbol, left, halt_state);
			output_machine(temp_machine, destination, alphabet);
			temp_machine=*this;
			temp_machine.add_move_transition(current, tape_symbol, right, halt_state);
			output_machine(temp_machine, destination, alphabet);
		}
	}

	//begin Owen Kellett modifications
	/**
	 * resets the machine to a blank tape and current state to starting state
	 * @post the machine's tape has been cleared, step count has been set to 0, and the current state is reset to the starting state
	 * @author Owen Kellett
	 */
	bool resetMachine()
	{
		//infinite_tape<T> newTape(tape.defaultValue());
		tape.reset();
		tape.reset_head();
		current = 0;
		steps = 0;
		run_status = run;
		return true;
	}

	/**
	 * get the mirror machine of *this.
	 * @return a turing machine identical to *this except with all left and right transitions reversed
	 * @author Owen Kellett
	 */
	turing_machine<T> reverseMachine()
	{
		turing_machine<T> reverse(transition_function.size(), 0);
		int j;

		for(j = 0; j < transition_function.size(); j++)
		{
			T write;
			typename infinite_tape<T>::direction move;
			typename std::vector<state<T> >::size_type nextState;

			if(transition_function[j].simulate_transition(0, write, move, nextState) == state<T>::run)
			{
				switch(move)
				{
					case infinite_tape<T>::left:
						reverse.add_move_transition(j,0, infinite_tape<T>::right, nextState);
						break;
					case infinite_tape<T>::right:
						reverse.add_move_transition(j,0, infinite_tape<T>::left, nextState);
						break;
					case infinite_tape<T>::do_nothing:
						reverse.add_write_transition(j,0, write, nextState);
				}
			}
			if(transition_function[j].simulate_transition(1, write, move, nextState) == state<T>::run)
			{
				switch(move)
				{
					case infinite_tape<T>::left:
						reverse.add_move_transition(j,1, infinite_tape<T>::right, nextState);
						break;
					case infinite_tape<T>::right:
						reverse.add_move_transition(j,1, infinite_tape<T>::left, nextState);
						break;
					case infinite_tape<T>::do_nothing:
						reverse.add_write_transition(j,1, write, nextState);
				}
			}
		}
		return reverse;
	}

	/**
	 * Rig the machine to a specified set of conditions
	 * Used for testing non-halt transformations, sets a portion of the tape equal to the elements in riggedTape; the read head is updated to the corresponding location of originalReadHead on the riggedTape; the desiredReadHead is used to set the desiredLocation iterator which is the hopeful location of the read head on the tape after the non-halt transformation is performed.
	 * @param riggedTape a vector of symbols representing a portion of a tape to which the machine should be set
	 * @param originalReadHead location on riggedTape of the read head at the start of the desired transformation
	 * @param desiredReadHead location on riggedTape of the read head at the end of the desired transformation
	 * @param state state to set the machine to
	 * @param leftBound reference to an iterator to set the leftBound of riggedTape when it is loaded on the tape
	 * @param rightBound reference to an iterator to set the rightBound of riggedTape when it is loaded on the tape
	 * @param desiredLocation reference to an iterator to set the location of riggedTape of the desired readhead after the transformation
	 * @author Owen Kellett
	 */
	void rigMachine(typename std::vector<T> &riggedTape,
			int originalReadHead,
			int desiredReadHead,
			typename std::vector<state<T> >::size_type state,
			typename std::list<T>::iterator &leftBound,
			typename std::list<T>::iterator &rightBound,
			typename std::list<T>::iterator &desiredLocation)
	{
		resetMachine();
		typename std::list<T>::iterator currentLocation = tape.currentLocation();
		desiredLocation = tape.currentLocation();
		leftBound = tape.currentLocation();

		for(int j = 0; j < riggedTape.size(); j++)
		{
			if(j == originalReadHead)
				tape.setCurrentLocation(currentLocation);
			*currentLocation = riggedTape[j];
			tape.checkIteratorRight(currentLocation);
			currentLocation++;
		}
		rightBound = currentLocation;
		rightBound--;
		if(desiredReadHead == -1)
			desiredLocation--;
		if(desiredReadHead == riggedTape.size())
			desiredLocation = currentLocation;

		current = state;
	}

	/**
	 * accessor to the states.
	 * @return reference to the set of states
	 * @author Owen Kellett
	 */
	std::vector<state<T> >& stateSet()
	{
		return transition_function;
	}

	/**
	 * accessor to the halt_state
	 * @return the index of the halt_state
	 * @author Owen Kellett
	 */
	transition_index_type haltState()
	{
		if(!halt_state_exists)
			return first_unreachable_state;
		return halt_state;
	}

	/**
	 * simulates the next move on the tape without actually running the tape.
	 * @return whether or not the machine halts
	 * @author Owen Kellett
	 */
	move_type nextMove()
	{
		T write;
		move_type move;
		typename std::vector<state<T> >::size_type nextState;
		if(transition_function[current].simulate_transition(tape.read(), write, move, nextState) == state<T>::halt)
			return halt_move;
		return move;
	}

	/**
	 * prints the turing machine in tmo (Turing Machine Owen) format.
	 * parsable by Owen's TM Simulator and also in valid dot format for creating visual graphs
	 * @post *this is printed in tmo format to out
	 * @author Owen Kellett
	 */
	void printTMO(std::ostream &out) const
	{
		out << "digraph TM" << std::endl;
		out << "{" << std::endl;
		out << "/*This machine represents " << number_represented() << " machines*/" << std::endl;
		out << "/*Settings*/" << std::endl;
		out << "rankdir = LR;" << std::endl;
		out << "start [shape=plaintext];" << std::endl;
		out << "halt [shape=plaintext];" << std::endl;

		typename std::vector<state<T> >::const_iterator i=transition_function.begin();
		int count = 0;
		T write;
		typename infinite_tape<T>::direction move;
		typename std::vector<state<T> >::size_type nextState;
		for(count = 0; i!=transition_function.end(); ++i)
		{
			if(count != halt_state)
				out << (count++) << " [shape=circle];" << std::endl;
		}

		out << std::endl;
		out << "/*GraphOutput*/" << std::endl;
		out << "/*NumStates = " << count << "*/" << std::endl;
		out << "/*Quadruple*/" << std::endl;
		if(implicitMachine())
			out << "/*Implicit*/" << std::endl;
		else
			out << "/*Explicit*/" << std::endl;
		out << "start -> 0;" << std::endl;

		i=transition_function.begin();
		count = 0;
		for(count = 0; i!=transition_function.end(); ++i, ++count)
		{
			if(count != halt_state)
			{
				if((*i).simulate_transition(0, write, move, nextState) == state<T>::run)
				{
					if(nextState == halt_state && write == 0 && move == infinite_tape<T>::do_nothing)
						out << count << " -> halt [label=\"" << 0 << "\"];" << std::endl;
					else
					{
						if(nextState == halt_state)
							out << count << " -> halt [label=\"" << 0 << ':';
						else
							out << count << " -> " << nextState << " [label=\"" << 0 << ':';
						if(write != 0)
							out << write << "\"];";
						else
						{
							if(move == infinite_tape<T>::left)
								out << "L\"];";
							else
								out << "R\"];";
						}
						out << std::endl;
					}
				}
				//else
					//out << count << " -> halt [label=\"" << 0 << "\"];" << std::endl;
				if((*i).simulate_transition(1, write, move, nextState) == state<T>::run)
				{
					if(nextState == halt_state && write == 1 && move == infinite_tape<T>::do_nothing)
						out << count << " -> halt [label=\"" << 1 << "\"];" << std::endl;
					else
					{
						if(nextState == halt_state)
							out << count << " -> halt [label=\"" << 1 << ':';
						else
							out << count << " -> " << nextState << " [label=\"" << 1 << ':';
						if(write != 1)
							out << write << "\"];";
						else
						{
							if(move == infinite_tape<T>::left)
								out << "L\"];";
							else
								out << "R\"];";
						}
						out << std::endl;
					}
				}
				//else
					//out << count << " -> halt [label=\"" << 1 << "\"];" << std::endl;
			}
		}
		out << "/*EndGraphOutput*/" << std::endl;
		out << "}" << std::endl;
	}

	/**
	 * Returns whether this machine is an implicit halt machine
	 * @return true if machine is implicit halt
	 * @author Owen Kellett
	 */
	bool implicitMachine() const
	{
		if(!halt_state_exists)
			return true;
		else
		{
			typename std::vector<state<T> >::const_iterator i=transition_function.begin();
			T write;
			typename infinite_tape<T>::direction move;
			typename std::vector<state<T> >::size_type nextState;
			for(i; i != transition_function.end(); i++)
			{
				if((*i).simulate_transition(0, write, move, nextState) == state<T>::run)
				{
					if(nextState == halt_state && (write != 0 || move != infinite_tape<T>::do_nothing))
						return false;
				}
				if((*i).simulate_transition(1, write, move, nextState) == state<T>::run)
				{
					if(nextState == halt_state && (write != 1 || move != infinite_tape<T>::do_nothing))
						return false;
				}
			}
			return true;
		}
	}

	//end Owen Kellett modifications

	private:

	template <typename OutputIterator>
	inline void output_machine(turing_machine<T>& to_output, OutputIterator destination) const
	{
		std::cout<<"fuck"<<std::endl;
	}

	template <typename OutputIterator>
	inline void output_machine(turing_machine<T>& to_output, OutputIterator destination, const std::list<T> &alphabet) const
	{
		#ifdef enable_optimisations
			if(run_checks(to_output, alphabet))
			{
				to_output.run_status=run;
				*(destination++)=to_output;
			}
		#else
			to_output.run_status=run;
			*(destination++)=to_output;
		#endif
	}

	public:
	private:

	/**
	 * run normalization checks
	 * @author Kyle Ross
	 */
	bool inline run_checks(const turing_machine<T> &to_check, const std::list<T> &alphabet) const
	{
		#ifdef first_write_optimisation
			T fw_write;
			typename infinite_tape<T>::direction fw_move;
			transition_index_type fw_next;
			to_check.transition_function[0].simulate_transition(0, fw_write, fw_move, fw_next);
			//if start state does not have 0:1 transition
			if(fw_write!=1 || fw_move!=do_nothing)
			{
				if(to_check.implicitMachine())
				{
					g_counter_implicit.increment(tm_stats::first_write_1, to_check.number_represented());
					c_counter_implicit.increment(tm_stats::first_write_1, 1);
				}
				g_counter_explicit.increment(tm_stats::first_write_1, to_check.number_represented_explicit());
				c_counter_explicit.increment(tm_stats::first_write_1, 1);
				return false;
			}

			#ifdef first_move_optimisation
				T fm_write;
				typename infinite_tape<T>::direction fm_move;
				transition_index_type fm_next;
				to_check.transition_function[fw_next].simulate_transition(1, fm_write, fm_move, fm_next);
				//if it's a move transition to the left ... if it's not a move, it must be an ss-transition or an ssprime-transition (i.e., 0:1 then 1:1=ss, 0:1 then 1:0=ssprime), so we don't handle that here
				if(fm_move==left)
				{
					if(to_check.implicitMachine())
					{
						g_counter_implicit.increment(tm_stats::first_move_right, to_check.number_represented());
						c_counter_implicit.increment(tm_stats::first_move_right, 1);
					}
					g_counter_explicit.increment(tm_stats::first_move_right, to_check.number_represented_explicit());
					c_counter_explicit.increment(tm_stats::first_move_right, 1);
					return false;
				}
			#endif
		#endif

		#ifdef no_ssprime_sprimes_optimisation
			//get information about current machine
			T c_read=to_check.tape.read(), c_write;
			typename infinite_tape<T>::direction c_move;
			transition_index_type c_current=to_check.current, c_next;

			//only need to check if we have a write-transition
			if(to_check.transition_function[c_current].simulate_transition(c_read, c_write, c_move, c_next)==run && c_move==do_nothing)
			{
				//check for the case where the currently-added transition forms the s:s' portion
				T n_read=c_write, n_write;
				typename infinite_tape<T>::direction n_move;
				transition_index_type n_current=c_next, n_next;

				//simulate transition and only continue check if a write-transition exists
				if(to_check.transition_function[n_current].simulate_transition(n_read, n_write, n_move, n_next)==run && n_move==do_nothing)
				{
					//read of first transition=write of second; write of first=read of second (trivially), so we have an c_read:c_write-c_write-c_read transition; eliminate
					if(c_read==n_write)
					{
						if(to_check.implicitMachine())
						{
							g_counter_implicit.increment(tm_stats::complex_nonproductive, to_check.number_represented());
							c_counter_implicit.increment(tm_stats::complex_nonproductive, 1);
						}
						g_counter_explicit.increment(tm_stats::complex_nonproductive, to_check.number_represented_explicit());
						c_counter_explicit.increment(tm_stats::complex_nonproductive, 1);
						return false;
					}
				}

				//check for the case where the currently-added transition forms the s':s portion
				T p_read=c_write, p_write;
				typename infinite_tape<T>::direction p_move;
				transition_index_type p_next;

				for(transition_index_type p_current=0, p_end=std::min(to_check.halt_state, to_check.first_unreachable_state); p_current!=p_end; ++p_current)
				{
					//remove a transition if -- given that the read is the current write -- it meets: 1)X exists 2)it is a quadruple write transition 3)the next state after the X transition is the current state 4)the symbol written is the same as the current read
					if(to_check.transition_function[p_current].simulate_transition(p_read, p_write, p_move, p_next)==run && p_move==do_nothing && p_next==c_current && p_write==c_read)
					{
						if(to_check.implicitMachine())
						{
							g_counter_implicit.increment(tm_stats::complex_nonproductive, to_check.number_represented());
							c_counter_implicit.increment(tm_stats::complex_nonproductive, 1);
						}
						g_counter_explicit.increment(tm_stats::complex_nonproductive, to_check.number_represented_explicit());
						c_counter_explicit.increment(tm_stats::complex_nonproductive, 1);
						return false;
					}
				}
			}
		#endif

		return true;
	}

	public:

protected:
	bool halt_state_exists;
	transition_index_type current, halt_state, first_unreachable_state;
	typename std::vector<state<T> > transition_function;
	infinite_tape<T> tape;
	transition_count_type steps;
	result_type run_status;
};



//displays a Turing machine
//preconditions: to_print is a valid Turing machine
//postconditions: to_print has been displayed according to the format specified by the print member function of the turing_machine class
//complexity: equal to that of the print member function of the turing_machine class
template <typename T>
std::ostream& operator<<(std::ostream &out, const turing_machine<T> &to_print)
{
	to_print.print(out);
	return out;
}

#endif
