/*Authored 2002    by  Kyle  Ross,  Undergraduate  Philosophy   Thesis
(Advisor: Professor Bram van Heuveln)

Modified 2003    by Owen Kellett, Undergraduate Research
(Advisor: Professor Bram van Heuveln)

This  file is part of the   "Generic Turing Machine Implementation for
Use in the Exploration of the Busy Beaver Problem".

The "Generic Turing Machine Implementation for  Use in the Exploration
of the Busy Beaver Problem" is free  software; you can redistribute it
and/or modify it under the terms of the  GNU General Public License as
published   by the Free Software Foundation;   either version 2 of the
License, or (at your option) any later version.

This programme is distributed in the hope that it  will be useful, but
WITHOUT    ANY  WARRANTY;  without   even  the    implied warranty  of
MERCHANTABILITY or FITNESS FOR  A  PARTICULAR  PURPOSE.  See  the  GNU
General Public License for more details.

You should have received   a copy of  the  GNU General Public  License
along  with  this   programme; if  not,  write   to the Free  Software
Foundation, Inc., 59  Temple Place, Suite  330, Boston, MA  02111-1307
USA (or visit http://www.fsf.org/licenses/licenses.html)

Created as part of the Busy Beaver Project:
Bram van Heuveln <heuveb@rpi.edu>
Boleshaw Szymanski <szymansk@cs.rpi.edu>
Selmer Bringsjord <selmer@rpi.edu>
Carlos Varela <cvarela@cs.rpi.edu>
Owen Kellett <kelleo@rpi.edu>
Shailesh Kelkar <kelkas@rpi.edu>
Kyle Ross <rossk2@rpi.edu>
Rensselaer Reasoning Group
Department of Cognitive Science
Department of Computer Science
Rensselaer Polytechnic Institute
Troy, New York 12180
U.S.A.
*/
//implements an infinite-length tape of an arbitrary alphabet to be used in a generalised turing machine implementation; assumes small-size alphabet (this could easily be expanded through a derived class)
//enforcing of alphabet rules (i.e. which symbols belong to the alphabet, etc.) will be left to the finite-state automaton

#include <fstream>
#include <iostream>
#include <iterator>
#include <list>
#include <vector>
#include <sys/stat.h>
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string>


#include "../bb/stats.hpp"
tm_stats g_counter_implicit;
tm_stats c_counter_implicit;
tm_stats g_counter_explicit;
tm_stats c_counter_explicit;

#include "../bb/infinite_tape.hpp"
#include "../bb/state.hpp"
#include "../bb/turing_machine.hpp"
#include "../bb/non_halt_detection.hpp"
#include "../bb/non_halt_data_structures.hpp"
#include "../bb/machine_storage.hpp"

#include "../bb/watch.hpp"
typedef unsigned short d_type;
enum machine_type{B, P, O, R};

void run_quadruple_bb_search(unsigned int size, std::list<turing_machine<d_type> > &machines_to_consider, const std::list<d_type> &alphabet, std::ostream &stats_stream, unsigned int transition_limit)
{
	char buffer[10];
	sprintf(buffer,"%d",size);
	std::string stateValue(buffer);
	std::string outerDir = "../output/BB" + stateValue;
	std::string holdoutDir = outerDir + "/holdouts";
	std::string bDir = outerDir + "/B" + stateValue;
	std::string oDir = outerDir + "/O" + stateValue;
	std::string pDir = outerDir + "/P" + stateValue;
	std::string rDir = outerDir + "/R" + stateValue;

	if(mkdir(outerDir.c_str(), 0744) == -1 ||
	  	mkdir(holdoutDir.c_str(), 0744) == -1 ||
		mkdir(bDir.c_str(), 0744) == -1 ||
		mkdir(oDir.c_str(), 0744) == -1 ||
		mkdir(pDir.c_str(), 0744) == -1 ||
		mkdir(rDir.c_str(), 0744) == -1)
	{
		std::cerr << "problem creating directories" << std::endl;
		return;
	}

	outputLog<d_type> loggingOutputHoldouts(holdoutDir + "/");
	championLog<d_type> loggingOutputB(bDir + "/");
	championLog<d_type> loggingOutputO(oDir + "/");
	championLog<d_type> loggingOutputP(pDir + "/");
	championLog<d_type> loggingOutputR(rDir + "/");

	//create statical recording variables
	unsigned int standard_count_implicit=0, nonstandard_count_implicit=0, holdout_count_implicit=0;
	unsigned int standard_count_explicit=0, nonstandard_count_explicit=0, holdout_count_explicit=0;

	unsigned int max_productivity_b=0, max_shifts_b=0, max_productivity_o=0, max_shifts_o=0;
	unsigned int max_productivity_p=0, max_shifts_p=0, max_productivity_r=0, max_shifts_r=0;

	//consider machines until machine number limit reached or all machines exhausted
	unsigned int machine_trial;
	for(machine_trial=0; machines_to_consider.size()!=0; ++machine_trial)
	{
		//if stats output is on and this is a thousand-mark machine, output stats
		if(machine_trial%100000==0)
		{
			#ifdef ross_debug_mode
				stats_stream<<"[stats-stream\n";
			#endif
			stats_stream<<"considering machine "<<machine_trial<<std::endl;
			stats_stream<<" holdouts="<<holdout_count_implicit<<std::endl;
			stats_stream<<" implicit standard="<<standard_count_implicit<<std::endl;
			stats_stream<<" implicit non-standard="<<nonstandard_count_implicit<<std::endl;
			stats_stream<<" explicit standard="<<standard_count_explicit<<std::endl;
			stats_stream<<" explicit non-standard="<<nonstandard_count_explicit<<std::endl;
			#ifdef ross_debug_mode
				stats_stream<<"stats-stream]\n";
			#endif
		}
		if(machine_trial%1000000==0)
		{
			stats_stream<<c_counter_implicit<<std::endl;
		}

		//get the next machine to be evaluated in DFS order
		turing_machine<d_type> current_machine=*(machines_to_consider.begin());
		machines_to_consider.pop_front();

		//initialise step counter
		unsigned int step_counter=0;


		if(backTrack(current_machine, 20))
		{
			g_counter_implicit.increment(tm_stats::backTracker, current_machine.number_represented());
			c_counter_implicit.increment(tm_stats::backTracker, 1);
			g_counter_explicit.increment(tm_stats::backTracker, current_machine.number_represented_explicit());
			c_counter_explicit.increment(tm_stats::backTracker, 1);
		}
		else if(loopStates(current_machine))
		{
			g_counter_implicit.increment(tm_stats::subsetLoop, current_machine.number_represented());
			c_counter_implicit.increment(tm_stats::subsetLoop, 1);
			g_counter_explicit.increment(tm_stats::subsetLoop, current_machine.number_represented_explicit());
			c_counter_explicit.increment(tm_stats::subsetLoop, 1);
		}
		else if(simpleLoopCheck(current_machine, 100))
		{
			g_counter_implicit.increment(tm_stats::simpleLoop, current_machine.number_represented());
			c_counter_implicit.increment(tm_stats::simpleLoop, 1);
			g_counter_explicit.increment(tm_stats::simpleLoop, current_machine.number_represented_explicit());
			c_counter_explicit.increment(tm_stats::simpleLoop, 1);
		}
		else if(multiSweepChristmasTree(current_machine, 1, 25000))
		{
			g_counter_implicit.increment(tm_stats::christmasTree, current_machine.number_represented());
			c_counter_implicit.increment(tm_stats::christmasTree, 1);
			g_counter_explicit.increment(tm_stats::christmasTree, current_machine.number_represented_explicit());
			c_counter_explicit.increment(tm_stats::christmasTree, 1);
		}
		else if(multiSweepChristmasTree(current_machine, 2, 50000))
		{
			g_counter_implicit.increment(tm_stats::alternateChristmasTree, current_machine.number_represented());
			c_counter_implicit.increment(tm_stats::alternateChristmasTree, 1);
			g_counter_explicit.increment(tm_stats::alternateChristmasTree, current_machine.number_represented_explicit());
			c_counter_explicit.increment(tm_stats::alternateChristmasTree, 1);
		}
		else if(leaningChristmasTree(current_machine, 25000))
		{
			g_counter_implicit.increment(tm_stats::leaningChristmasTree, current_machine.number_represented());
			c_counter_implicit.increment(tm_stats::leaningChristmasTree, 1);
			g_counter_explicit.increment(tm_stats::leaningChristmasTree, current_machine.number_represented_explicit());
			c_counter_explicit.increment(tm_stats::leaningChristmasTree, 1);
		}
		else if(multiSweepChristmasTree(current_machine, 3, 100000))
		{
			g_counter_implicit.increment(tm_stats::tripleSweepChristmasTree, current_machine.number_represented());
			c_counter_implicit.increment(tm_stats::tripleSweepChristmasTree, 1);
			g_counter_explicit.increment(tm_stats::tripleSweepChristmasTree, current_machine.number_represented_explicit());
			c_counter_explicit.increment(tm_stats::tripleSweepChristmasTree, 1);
		}
		else if(multiSweepChristmasTree(current_machine, 4, 100000))
		{
			g_counter_implicit.increment(tm_stats::quadrupleSweepChristmasTree, current_machine.number_represented());
			c_counter_implicit.increment(tm_stats::quadrupleSweepChristmasTree, 1);
			g_counter_explicit.increment(tm_stats::quadrupleSweepChristmasTree, current_machine.number_represented_explicit());
			c_counter_explicit.increment(tm_stats::quadrupleSweepChristmasTree, 1);
		}
		else if(multiSweepChristmasTree(current_machine, 5, 100000))
		{
			g_counter_implicit.increment(tm_stats::quintupleSweepChristmasTree, current_machine.number_represented());
			c_counter_implicit.increment(tm_stats::quintupleSweepChristmasTree, 1);
			g_counter_explicit.increment(tm_stats::quintupleSweepChristmasTree, current_machine.number_represented_explicit());
			c_counter_explicit.increment(tm_stats::quintupleSweepChristmasTree, 1);
		}
		else if(counter(current_machine, 50000))
		{
			g_counter_implicit.increment(tm_stats::counter, current_machine.number_represented());
			c_counter_implicit.increment(tm_stats::counter, 1);
			g_counter_explicit.increment(tm_stats::counter, current_machine.number_represented_explicit());
			c_counter_explicit.increment(tm_stats::counter, 1);
		}

		else
		{
		current_machine.resetMachine();
		//run the machine until it halts or the step-limit has been reached
		#ifdef enable_optimisations
			#ifdef blank_tape_optimisation
				while(step_counter<transition_limit && current_machine.run_one()!=turing_machine<d_type>::halt && !(step_counter>0 && current_machine.blank_tape()))
			#else
				while(step_counter<transition_limit && current_machine.run_one()!=turing_machine<d_type>::halt)
			#endif
		#else
			while(step_counter<transition_limit && current_machine.run_one()!=turing_machine<d_type>::halt)
		#endif
		{
			++step_counter;
		}

		#ifdef enable_optimisations
			#ifdef blank_tape_optimisation
				if(step_counter>0 && current_machine.blank_tape())
				{
					#ifdef ross_debug_mode
						std::cout<<"MACHINE ELIMINATED BECAUSE IT IS A BLANK-TAPE MACHINE"<<std::endl;
						std::cout<<current_machine<<std::endl;
					#endif
					if(current_machine.implicitMachine())
					{
						g_counter_implicit.increment(tm_stats::empty_tape, current_machine.number_represented());
						c_counter_implicit.increment(tm_stats::empty_tape, 1);
					}
					g_counter_explicit.increment(tm_stats::empty_tape, current_machine.number_represented_explicit());
					c_counter_explicit.increment(tm_stats::empty_tape, 1);
				}
				else
			#endif
		#endif

		//if(subsetLoop);
		//if the machine is still running, add it as an holdout
		if(current_machine.get_status()==turing_machine<d_type>::run)
		{
			if(simpleLoopCheck(current_machine, 1000))
			{
				g_counter_implicit.increment(tm_stats::simpleLoop, current_machine.number_represented());
				c_counter_implicit.increment(tm_stats::simpleLoop, 1);
				g_counter_explicit.increment(tm_stats::simpleLoop, current_machine.number_represented_explicit());
				c_counter_explicit.increment(tm_stats::simpleLoop, 1);
			}
			else
			{
				g_counter_implicit.increment(tm_stats::holdout, current_machine.number_represented());
				c_counter_implicit.increment(tm_stats::holdout, 1);
				g_counter_explicit.increment(tm_stats::holdout, current_machine.number_represented_explicit());
				c_counter_explicit.increment(tm_stats::holdout, 1);

				loggingOutputHoldouts.print(current_machine);

				++holdout_count_implicit;
				++holdout_count_explicit;
			}
		}
		//otherwise, it's a non-standard halter xor a standard halter
		else if(current_machine.in_halt_state())
		{
			//if it's a non-standard halter:
			if(current_machine.implicitMachine())
			{
				if(current_machine.in_standard_position())
				{
					g_counter_implicit.increment(tm_stats::candidate, current_machine.number_represented());
					c_counter_implicit.increment(tm_stats::candidate, 1);

					++standard_count_implicit;

					//is it a new most-shift machine?
					if(current_machine.get_steps_count()>max_shifts_b)
					{
						max_shifts_b=current_machine.get_steps_count();
						loggingOutputB.printChampion(current_machine, championLog<d_type>::transitions, max_shifts_b);
						stats_stream<<"a new b champion found with shift count "<<max_shifts_b<<std::endl;
					}

					//is it a new most-productive machine?
					if(current_machine.count_occurrences(1)>=max_productivity_b)
					{
						max_productivity_b=current_machine.count_occurrences(1);
						loggingOutputB.printChampion(current_machine, championLog<d_type>::ones, max_productivity_b);
						stats_stream<<"a new B champion found with productivity "<<max_productivity_b<<std::endl;
					}
				}
				else
				{
					g_counter_implicit.increment(tm_stats::not_in_standard_position, current_machine.number_represented());
					c_counter_implicit.increment(tm_stats::not_in_standard_position, 1);

					++nonstandard_count_implicit;
				}
				//is it a new most-shift machine?
				if(current_machine.get_steps_count()>max_shifts_o)
				{
					max_shifts_o=current_machine.get_steps_count();
					loggingOutputO.printChampion(current_machine, championLog<d_type>::transitions, max_shifts_o);
					stats_stream<<"a new o champion found with shift count "<<max_shifts_o<<std::endl;
				}

				//is it a new most-productive machine?
				if(current_machine.count_occurrences(1)>=max_productivity_o)
				{
					max_productivity_o=current_machine.count_occurrences(1);
					loggingOutputO.printChampion(current_machine, championLog<d_type>::ones, max_productivity_o);
					stats_stream<<"a new O champion found with productivity "<<max_productivity_o<<std::endl;
				}
			}

			if(current_machine.in_standard_position())
			{
				g_counter_explicit.increment(tm_stats::candidate, current_machine.number_represented_explicit());
				c_counter_explicit.increment(tm_stats::candidate, 1);

				++standard_count_explicit;

				//is it a new most-shift machine?
				if(current_machine.get_steps_count()>max_shifts_p && !current_machine.implicitMachine())
				{
					max_shifts_p=current_machine.get_steps_count();
					loggingOutputP.printChampion(current_machine, championLog<d_type>::transitions, max_shifts_p);
					stats_stream<<"a new p champion found with shift count "<<max_shifts_p<<std::endl;
				}

				//is it a new most-productive machine?
				if(current_machine.count_occurrences(1)>=max_productivity_p && !current_machine.implicitMachine())
				{
					max_productivity_p=current_machine.count_occurrences(1);
					loggingOutputP.printChampion(current_machine, championLog<d_type>::ones, max_productivity_p);
					stats_stream<<"a new P champion found with productivity "<<max_productivity_p<<std::endl;
				}
			}
			else
			{
				g_counter_explicit.increment(tm_stats::not_in_standard_position, current_machine.number_represented_explicit());
				c_counter_explicit.increment(tm_stats::not_in_standard_position, 1);

				++nonstandard_count_explicit;
			}
			//is it a new most-shift machine?
			if(current_machine.get_steps_count()>max_shifts_r && !current_machine.implicitMachine())
			{
				max_shifts_r=current_machine.get_steps_count();
				loggingOutputR.printChampion(current_machine, championLog<d_type>::transitions, max_shifts_r);
				stats_stream<<"a new r champion found with shift count "<<max_shifts_r<<std::endl;
			}

			//is it a new most-productive machine?
			if(current_machine.count_occurrences(1)>=max_productivity_r && !current_machine.implicitMachine())
			{
				max_productivity_r=current_machine.count_occurrences(1);
				loggingOutputR.printChampion(current_machine, championLog<d_type>::ones, max_productivity_r);
				stats_stream<<"a new R champion found with productivity "<<max_productivity_r<<std::endl;
			}
		}
		//halted because transition does not exist
		else
		{
			//generate appropriate children (i.e. taking into account implicit/explicit halt)
			/*if(implicit)
			{
				current_machine.generate_tn_quadruple_children(alphabet, front_inserter(machines_to_consider));
			}
			else
			{
				current_machine.generate_tn_quadruple_children_explicit(alphabet, front_inserter(machines_to_consider));
			}*/
			current_machine.generate_all_tn_quadruple_children(alphabet, front_inserter(machines_to_consider));
		}
		}
	}

	stats_stream<<"<finished>"<<std::endl;
	stats_stream<<"considered machine "<<machine_trial<<std::endl;
	stats_stream<<" holdouts="<<holdout_count_implicit<<std::endl;
	stats_stream<<" implicit standard="<<standard_count_implicit<<std::endl;
	stats_stream<<" implicit non-standard="<<nonstandard_count_implicit<<std::endl;
	stats_stream<<" explicit standard="<<standard_count_explicit<<std::endl;
	stats_stream<<" explicit non-standard="<<nonstandard_count_explicit<<std::endl;
}


void run_simple_quadruple(unsigned int size, unsigned int halt_limit, std::ostream &cand_file)
{
	cand_file<<"running ";
	cand_file<<"("<<size<<") with step-limit of "<<halt_limit<<std::endl;
	display_optimisation_settings(cand_file);

	typedef turing_machine<unsigned short> m_type;

	//create the initial machine queue
	std::list<m_type> machines;
	m_type original_machine(size+1, 0); //the root machine of the tree
	original_machine.set_halt_state(size);
	machines.push_front(original_machine);
	std::cerr<<"implicit target="<<original_machine.number_represented()<<std::endl;
	std::cerr<<"explicit target="<<original_machine.number_represented_explicit()<<std::endl;

	//create the alphabet
	std::list<unsigned short> alphabet;
	alphabet.push_back(0);
	alphabet.push_back(1);

	//create a list for the candidates
	std::list<m_type> bs_list;

	run_quadruple_bb_search(
		size,
		machines, //list of machines
		alphabet, //1-0 alphabet
		cand_file, //stats output
		halt_limit); //number of transitions to run
}


int main(int argc, char *argv[])
{
	if(argc != 3)
	{
		std::cerr << "Wrong number of arguments" << std::endl;
		return 0;
	}
	int numStates = atoi(argv[1]);
	int haltLimit = atoi(argv[2]);
	if(numStates < 0 || haltLimit < 0)
	{
		std::cerr << "Invalid arguments" << std::endl;
		return 0;
	}

	run_simple_quadruple(numStates, haltLimit, std::cout);

	std::cerr<<"observed "<<c_counter_implicit<<std::endl;
	std::cerr<<"representative "<<g_counter_implicit<<std::endl;
	std::cerr<<"observed "<<c_counter_explicit<<std::endl;
	std::cerr<<"representative "<<g_counter_explicit<<std::endl;

	return EXIT_SUCCESS;
}



